#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/pgtable.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <asm/highmem.h>
#include <linux/rmap.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

MODULE_DESCRIPTION("simple mmap driver");
MODULE_AUTHOR("PSO");
MODULE_LICENSE("Dual BSD/GPL");

#define MY_MAJOR	42
/* how many pages do we actually kmalloc */
#define NPAGES		16

/* character device basic structure */
static struct cdev mmap_cdev;

/* pointer to kmalloc'd area */
static void *kmalloc_ptr;

/* pointer to the kmalloc'd area, rounded up to a page boundary */
static char *kmalloc_area;

static int my_open(struct inode *inode, struct file *filp)
{
	return 0;
}

static int my_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static int my_mmap(struct file *filp, struct vm_area_struct *vma)
{
	int ret;
	long length = vma->vm_end - vma->vm_start;

	/* Do not map more than we can. */
	if (length > NPAGES * PAGE_SIZE)
		return -EIO;

	/* Map the whole physically contiguous area in one piece. */
	ret = remap_pfn_range(vma, vma->vm_start,
			virt_to_phys((void *)kmalloc_area) >> PAGE_SHIFT,
			length, vma->vm_page_prot);
	if (ret < 0) {
		printk(KERN_ERR "could not map address area\n");
		return ret;
	}

	return 0;
}

static const struct file_operations mmap_fops = {
	.owner = THIS_MODULE,
	.open = my_open,
	.release = my_release,
	.mmap = my_mmap
};

static int __init my_mmap_init(void)
{
	int ret = 0;
	int i;

	ret = register_chrdev_region(MKDEV(MY_MAJOR, 0), 1, "mymmap");
	if (ret < 0) {
		printk(KERN_ERR "could not register region\n");
		goto out;
	}

	/* Allocate NPAGES+2 pages using kmalloc */
	kmalloc_ptr = kmalloc((NPAGES + 2) * PAGE_SIZE, GFP_KERNEL);
	if (kmalloc_ptr == NULL) {
		ret = -ENOMEM;
		printk(KERN_ERR "could not allocate memory\n");
		goto out_unreg;
	}

	/* Round kmalloc_ptr to nearest page start address. */
	kmalloc_area = (char *) PAGE_ALIGN(((unsigned long)kmalloc_ptr));

	/* Mark pages as reserved. */
	for (i = 0; i < NPAGES * PAGE_SIZE; i += PAGE_SIZE)
		SetPageReserved(virt_to_page(((unsigned long)kmalloc_area)+i));

	/* Write data in to be mapped area. */
	memset(kmalloc_area, 'a', NPAGES * PAGE_SIZE);

	/* Init device. */
	cdev_init(&mmap_cdev, &mmap_fops);
	ret = cdev_add(&mmap_cdev, MKDEV(MY_MAJOR, 0), 1);
	if (ret < 0) {
		printk(KERN_ERR "could not add device\n");
		goto out_kfree;
	}

	return 0;

out_kfree:
	kfree(kmalloc_ptr);
out_unreg:
	unregister_chrdev_region(MKDEV(MY_MAJOR, 0), 1);
out:
	return ret;
}

static void __exit my_mmap_exit(void)
{
	int i;

	cdev_del(&mmap_cdev);

	/* Clear reservation on pages and free mem. */
	for (i = 0; i < NPAGES * PAGE_SIZE; i += PAGE_SIZE)
		ClearPageReserved(virt_to_page(((unsigned long)kmalloc_area)+i));
	kfree(kmalloc_ptr);

	unregister_chrdev_region(MKDEV(MY_MAJOR, 0), 1);
}

module_init(my_mmap_init);
module_exit(my_mmap_exit);
